from pathlib import Path
import numpy as np

home = Path('~/mci2ad').expanduser()

train_path = home / 'npy/train_ad_cn_2mm'
val_path = home / 'npy/val_ad_cn_2mm'
labels_train = np.load(train_path / "labels.npy")
labels_val = np.load(val_path / "labels.npy")
len(labels_train), len(labels_val)

# First cross-val - complete copy
xval = '04'
switch_start = 194
switch_total = 97

train_dest_path = Path(f"npy/train_ad_cn_2mm_{xval}")
val_dest_path = Path(f"npy/val_ad_cn_2mm_{xval}")

train_dest_labels = train_dest_path / "labels.npy"
val_dest_labels = val_dest_path / "labels.npy"

labels_train_x = labels_train.copy()
labels_val_x = labels_val.copy()

train_switch_list = np.zeros(len(labels_train))
val_switch_list = np.zeros(len(labels_val))
if switch_total > 0:
    train_switch_list[switch_start:switch_start+switch_total] = 1
    val_switch_list[:switch_total] = 1
    labels_train_x[switch_start:switch_start+switch_total] = labels_val[:switch_total]
    labels_val_x[:switch_total] = labels_train[switch_start:switch_start+switch_total]
j = 0
for i,switch in enumerate(train_switch_list):
    suffix = f"{i:04d}"
    suffix_val = f"{j:04d}"
    dest_train_base = train_dest_path / f"X_base.{suffix}.npy"
    dest_train_long = train_dest_path / f"X_long.{suffix}.npy"
    if switch:
        source_base = val_path / f"X_base.{suffix_val}.npy"
        source_long = val_path / f"X_long.{suffix_val}.npy"
        j += 1
    else:
        source_base = train_path / f"X_base.{suffix}.npy"
        source_long = train_path / f"X_long.{suffix}.npy"

    dest_train_base.symlink_to(source_base)
    dest_train_long.symlink_to(source_long)
    print(f"Links: {dest_train_base} -> {source_base}, {dest_train_long} -> {source_long}")

for i,switch in enumerate(val_switch_list):
    suffix_val = f"{i:04d}"
    suffix = f"{i+switch_start:04d}"
    dest_val_base = val_dest_path / f"X_base.{suffix_val}.npy"
    dest_val_long = val_dest_path / f"X_long.{suffix_val}.npy"
    if switch:
        source_base = train_path / f"X_base.{suffix}.npy"
        source_long = train_path / f"X_long.{suffix}.npy"
    else:
        source_base = val_path / f"X_base.{suffix_val}.npy"
        source_long = val_path / f"X_long.{suffix_val}.npy"

    dest_val_base.symlink_to(source_base)
    dest_val_long.symlink_to(source_long)

    print(f"Links: {dest_val_base} -> {source_base}, {dest_val_long} -> {source_long}")

np.save(train_dest_labels, labels_train_x)
np.save(val_dest_labels, labels_val_x)
