import numpy as np
from pathlib import Path
import logging

class DataSaver:

    top = Path("npy")
    
    def __init__(self, name, X_base, X_long, y):
        if len(X_base) != len(X_long):
            logging.error(f"X_base length is {len(X_base)} while X_long length is {len(X_long)}")
            raise Exception("Length mismatch")
        if len(X_base) != len(y):
            logging.error(f"X_base ength is {len(X_base)} while y length is {len(y)}")
            raise Exception("Length mismatch")
        self.name = name
        self.X_base = X_base
        self.X_long = X_long
        self.y = y
        self.save()

    def set_directory(self):
        self.dir = self.top / self.name
        if not self.dir.is_dir():
            self.dir.mkdir()

    def save(self):
        self.set_directory()
        for i in range(len(self.X_base)):
            logging.info(f"#{i} saving data...")
            suffix = f"{i:04d}"
            X_base_path = self.dir / f"X_base.{suffix}.npy"
            X_long_path = self.dir / f"X_long.{suffix}.npy"
            if X_base_path.exists():
                X_base_path.unlink()
            if X_long_path.exists():
                X_long_path.unlink()
            np.save(X_base_path, self.X_base[i])
            logging.info("Baseline saved")
            np.save(X_long_path, self.X_long[i])
            logging.info("Longitudinal saved")
        y_path = self.dir / f"labels.npy"
        if y_path.exists():
            y_path.unlink()
        np.save(y_path, self.y)
        logging.info("Labels saved")

class DataSaverBase(DataSaver):

    def __init__(self, name, X_base, y):
        if len(X_base) != len(y):
            logging.error(f"X_base ength is {len(X_base)} while y length is {len(y)}")
            raise Exception("Length mismatch")
        self.name = name
        self.X_base = X_base
        self.y = y
        self.save()

    def save(self):
        self.set_directory()
        for i in range(len(self.X_base)):
            logging.info(f"#{i} saving data...")
            suffix = f"{i:04d}"
            X_base_path = self.dir / f"X_base.{suffix}.npy"
            if X_base_path.exists():
                X_base_path.unlink()
            np.save(X_base_path, self.X_base[i])
            logging.info("Baseline saved")
        y_path = self.dir / f"labels.npy"
        if y_path.exists():
            y_path.unlink()
        np.save(y_path, self.y)
        logging.info("Labels saved")

