import numpy as np
from pathlib import Path
import nibabel as nib
import logging
from sklearn.model_selection import train_test_split
from make_data import Pre_Data_Volume
import json
import pandas as pd
from data_saver import DataSaver

logging.getLogger().setLevel(logging.INFO)

NIFTI_SHAPE = (91, 109, 91)
MAX_LOAD = 20000
limit = -1

stable_df = pd.read_csv("stable_adni1_longitudinal.csv")
progressive_df = pd.read_csv("progressive_adni1_longitudinal.csv")
all_df = stable_df.append(progressive_df)
stable_pts = list(stable_df["Subject ID"])
progressive_pts = list(progressive_df["Subject ID"])

if len(stable_pts) > MAX_LOAD:
    stable_pts = stable_pts[:MAX_LOAD]

if len(progressive_pts) > MAX_LOAD:
    progressive_pts = progressive_pts[:MAX_LOAD]

all_pts = stable_pts + progressive_pts
labels = [0] * len(stable_pts) + [1] * len(progressive_pts)

pts_train_and_val, pts_test, labels_train_and_val, labels_test = train_test_split(all_pts, labels, random_state=42, stratify=labels, test_size=0.25)

logging.info(f"{len(stable_pts)} {stable_pts[0]}" )
logging.info(f"{len(progressive_pts)} {progressive_pts[0]}")

pts_train, pts_val, labels_train, labels_val = train_test_split(pts_train_and_val, labels_train_and_val, random_state=43, stratify=labels_train_and_val, test_size=0.25)

baseline_ids_train = []
longitudinal_ids_train = []
for subject in pts_train:
    baseline_ids_train.append(all_df[all_df["Subject ID"] == subject].iloc[0]["Baseline ID"])
    longitudinal_ids_train.append(all_df[all_df["Subject ID"] == subject].iloc[0]["Longitudinal ID"])

baseline_ids_val = []
longitudinal_ids_val = []
for subject in pts_val:
    baseline_ids_val.append(all_df[all_df["Subject ID"] == subject].iloc[0]["Baseline ID"])
    longitudinal_ids_val.append(all_df[all_df["Subject ID"] == subject].iloc[0]["Longitudinal ID"])

limit_base_train = limit if limit>0 else len(baseline_ids_train)
limit_long_train = limit if limit>0 else len(longitudinal_ids_train)
limit_total_train = limit if limit>0 else len(labels_train)
limit_base_val = limit if limit>0 else len(baseline_ids_val)
limit_long_val = limit if limit>0 else len(longitudinal_ids_val)
limit_total_val = limit if limit>0 else len(labels_val)

data_train = Pre_Data_Volume(NIFTI_SHAPE, 
                            baseline_ids_train[:limit_base_train], 
                            longitudinal_ids_train[:limit_long_train])
data_val = Pre_Data_Volume(NIFTI_SHAPE, 
                            baseline_ids_val[:limit_base_val], 
                            longitudinal_ids_val[:limit_long_val])

logging.info("saving files...")

train_saver = DataSaver("train_adni1_2mm", data_train.X_base, data_train.X_long, labels_train[:limit_total_train])
val_saver = DataSaver("val_adni1_2mm", data_val.X_base, data_val.X_long, labels_val[:limit_total_val])

logging.info("All saves complete")