import numpy as np
from pathlib import Path
import nibabel as nib
import logging
from sklearn.model_selection import train_test_split
from make_data import Pre_Data_Volume
import json
import pandas as pd
from data_saver import DataSaver

logging.getLogger().setLevel(logging.INFO)

NIFTI_SHAPE = (91, 109, 91)
MAX_LOAD = 20000
limit = -1

ad_cn = pd.read_csv("inputs/ad_cn_long.csv")

if len(ad_cn) > MAX_LOAD:
    ad_cn = ad_cn[:MAX_LOAD]

index = np.array(ad_cn.index)
all_imgs = np.array(ad_cn["Image_ID base"])
all_imgs_long = np.array(ad_cn["Image_ID long"])
labels = np.array(ad_cn["Is_AD"])

index_tn, index_val, labels_train, labels_val = train_test_split(index, labels, random_state=43, stratify=labels, test_size=0.25)

baseline_ids_train = np.zeros(len(index_tn), dtype=int)
long_ids_train = np.zeros(len(index_tn), dtype=int)
for i, ix in enumerate(index_tn):
    baseline_ids_train[i] = all_imgs[ix]
    long_ids_train[i] = all_imgs_long[ix]

baseline_ids_val = np.zeros(len(index_val), dtype=int)
long_ids_val = np.zeros(len(index_val), dtype=int)
for i, ix in enumerate(index_val):
    baseline_ids_val[i] = all_imgs[ix]
    long_ids_val[i] = all_imgs_long[ix]

limit_base_train = limit if limit>0 else len(baseline_ids_train)
limit_long_train = limit if limit>0 else len(long_ids_train)
limit_total_train = limit if limit>0 else len(labels_train)
limit_base_val = limit if limit>0 else len(baseline_ids_val)
limit_long_val = limit if limit>0 else len(long_ids_val)
limit_total_val = limit if limit>0 else len(labels_val)

data_train = Pre_Data_Volume(NIFTI_SHAPE, 
                            baseline_ids_train[:limit_base_train], long_ids_train[:limit_long_train]) 
data_val = Pre_Data_Volume(NIFTI_SHAPE, 
                            baseline_ids_val[:limit_base_val], long_ids_val[:limit_long_val])

logging.info("saving files...")

train_saver = DataSaver("train_ad_cn_2mm", data_train.X_base, data_train.X_long, labels_train[:limit_total_train])
val_saver = DataSaver("val_ad_cn_2mm", data_val.X_base, data_val.X_long, labels_val[:limit_total_val])

logging.info("All saves complete")
