from pathlib import Path
import nibabel as nib
import numpy as np
import logging
from typing import Iterable

logging.getLogger().setLevel(logging.INFO)

brains: Iterable[Path] = Path("ADNI").rglob("*brain_reg2.nii.gz")

for i, brain in enumerate(brains):
    norm_path: Path = brain.parent / 'brain_reg2_norm.nii.gz'
    if norm_path.exists():
        logging.info(f"Already normalized #{i}: {norm_path}")
        continue
    brain_img: nib.Nifti1Image = nib.load(brain)
    brain_vol: np.ndarray = brain_img.get_fdata(dtype=np.float32)
    logging.info(f"Loaded brain #{i}: {str(brain)}")
    max_int  = brain_vol.max()
    logging.info(f"max_int: {max_int}")
    min_int = brain_vol.min()
    logging.info(f"min_int: {min_int}")
    norm_vol = (brain_vol - min_int) * 1000 / (max_int - min_int)
    nib.save(nib.Nifti1Image(norm_vol, brain_img.affine), norm_path)
    logging.info(f"Saved normalized brain: #{i}: {norm_path}")