import numpy as np
from pathlib import Path
import nibabel as nib
import logging
from sklearn.model_selection import train_test_split
from make_data import Pre_Data_Volume
import json
import pandas as pd
from data_saver import DataSaver

logging.getLogger().setLevel(logging.INFO)

NIFTI_SHAPE = (91, 109, 91)
MAX_LOAD = 20000
limit = -1

stable_df = pd.read_csv("inputs/stable_longitudinal.csv")
progressive_df = pd.read_csv("inputs/progressive_longitudinal.csv")
all_df = stable_df.append(progressive_df)
stable_pts = list(stable_df["Subject ID"])
progressive_pts = list(progressive_df["Subject ID"])

if len(stable_pts) > MAX_LOAD:
    stable_pts = stable_pts[:MAX_LOAD]

if len(progressive_pts) > MAX_LOAD:
    progressive_pts = progressive_pts[:MAX_LOAD]

all_pts = stable_pts + progressive_pts
labels = [0] * len(stable_pts) + [1] * len(progressive_pts)

pts_train_and_val, pts_test, labels_train_and_val, labels_test = train_test_split(all_pts, labels, random_state=42, stratify=labels, test_size=0.25)

logging.info(f"{len(stable_pts)} {stable_pts[0]}" )
logging.info(f"{len(progressive_pts)} {progressive_pts[0]}")

baseline_ids_test = []
longitudinal_ids_test = []
for subject in pts_test:
    baseline_ids_test.append(all_df[all_df["Subject ID"] == subject].iloc[0]["Baseline ID"])
    longitudinal_ids_test.append(all_df[all_df["Subject ID"] == subject].iloc[0]["Longitudinal ID"])

limit_base_test = limit if limit>0 else len(baseline_ids_test)
limit_long_test = limit if limit>0 else len(longitudinal_ids_test)
limit_total_test = limit if limit>0 else len(labels_test)

data_test = Pre_Data_Volume(NIFTI_SHAPE, 
                            baseline_ids_test[:limit_base_test], 
                            longitudinal_ids_test[:limit_long_test])

logging.info("saving files...")

train_saver = DataSaver("test_2mm", data_test.X_base, data_test.X_long, labels_test[:limit_total_test])

logging.info("All saves complete")