import numpy as np
from importlib import reload
import pandas as pd
import json
from os.path import isfile
from pathlib import Path
import logging
import tensorflow as tf
from tensorflow.keras.applications import VGG16, ResNet50V2, NASNetLarge
from tensorflow.keras.layers import BatchNormalization, Conv3D, ZeroPadding3D, MaxPooling3D, \
    GlobalAveragePooling3D, AveragePooling3D, Activation, LeakyReLU, Dense, Flatten, Input, \
    add, subtract, LeakyReLU, Dropout, concatenate, MaxPool3D, multiply, SpatialDropout3D, \
    GlobalMaxPooling3D
from tensorflow.keras import Model
from tensorflow.keras.regularizers import l2
from tensorflow.keras import backend as K
from tensorflow.keras.optimizers import SGD, Optimizer, Adam, RMSprop, Adagrad
from tensorflow.keras.callbacks import ModelCheckpoint, CSVLogger, EarlyStopping, LearningRateScheduler
from tensorflow.keras.utils import plot_model, Sequence
from sklearn.metrics import classification_report, confusion_matrix, balanced_accuracy_score, roc_auc_score
import callbacks
reload(callbacks)
PlotCallback = callbacks.PlotCallback
from make_data import End_Data_Volume
from scipy.ndimage import rotate, zoom
Dataset = tf.data.Dataset

config = tf.compat.v1.ConfigProto()

# Don't pre-allocate memory; allocate as-needed
config.gpu_options.allow_growth = True

class BaseModel:

    top = Path("npy")
    limit = -1
    is_siamese = False

    def __init__(self, name='base_model', comment="", compile_now=True):
        self.name = name
        self.comment = comment
        self.attempt = 0
        self.input_shape = [91, 109, 91, 1]
        self.epochs = 160
        self.train_dir = 'train'
        self.val_dir = 'val'
        self.test_dir = 'test'
        self.xval = 0
        self.single_dir_label = 'base'
        self.classes = 1
        self.model: Model
        self.child_model: Model
        self.child_model2 = None
        self.val: list
        self.pred: list
        self.n: int
        self.batch_size: 10
        self.stages: tuple
        self.filters:tuple
        self.child_model: Model
        self.learning_rate: int
        self.learning_rate = 1e-4
        self.lr_end = 1e-4
        self.lr_start = 1e-4
        self.opt = Adam(1e-4,epsilon=1e-8)
        self.tail_name = "medium_1300"
        self.tail_args = []
        self.child_transfer = ""
        self.prev_checkpoint = ""
        self.tails = {
            "short_128": Tails.short_128,
            "medium_1300": Tails.medium_1300,
            "long_2048": Tails.long_2048,
            "gap": Tails.gap,
            "gmp": Tails.gmp,
            "flat": Tails.flat,
            "flat128": Tails.flat128}
        self.parameters = {}
        self.results = {}
        self.freeze_child = False
        self.unfreeze_end = False
        self.freeze_tail = False
        self.patience = 30
        self.compile_now = compile_now
        self.set_parameters()
        if self.compile_now:
            self.setup_model()

    def setup_model(self):
        self.name = self.available_name(self.name)
        logging.info("\n\n**************************************")
        logging.info("Model instantiated")
        logging.info(f"Base class: {type(self).__name__}")
        logging.info(f"Name: {self.name}")
        logging.info(f"Comment: {self.comment}")
        logging.info(f"Train dir: {self.train_dir}")
        logging.info(f"Val dir: {self.val_dir}")
        self.logger = CSVLogger(f"plots/{self.name}_log.csv", append=True)
        self.plot = PlotCallback(self.name)
        self.build_model()
        self.load_data()
        self.compile_model()

    def set_parameters(self):
        pass

    def available_name(self, name):
        for i in range(1,999):
            suffix = f"{i:03d}"
            xval_suffix = self.xval_suffix()
            possible_name = f"{name}{xval_suffix}_{suffix}"
            logpath = Path('logs')
            plotpath = Path('plots')
            if not (plotpath / f"{possible_name}_log.csv").exists() \
                and not (plotpath / f"{possible_name}_plot.csv").exists() \
                and not (logpath / f"{possible_name}_results.json").exists():
                return possible_name
        logging.error(f"No available filesnames for {name}_xxx from 001 to 999")
        raise BaseException("No available filenames")

    def xval_suffix(self):
        return "" if self.xval == 0 else f"_{self.xval:02d}"

    def load_data(self):
        suffix = self.xval_suffix()
        self.train_path = self.top / (self.train_dir + suffix)
        self.val_path = self.top / (self.val_dir + suffix)
        self.test_path = self.top / self.test_dir
        self.load_labels()
        self.verify_image_quantity()
        self.define_datasets()

    def load_labels(self):
        logging.info("Loading labels...")
        logging.info(f"Train path: {self.train_path}")
        logging.info(f"Validation path: {self.val_path}")
        logging.info(f"Test path: {self.test_path}")
        self.y_train = np.load(self.train_path / 'labels.npy')
        self.y_val = np.load(self.val_path / 'labels.npy')
        self.y_test = np.load(self.test_path / 'labels.npy')
        if self.limit > 0:
            self.y_train = self.y_train[:self.limit]
            self.y_val = self.y_val[:self.limit]
            self.y_test = self.y_test[:self.limit]
        self.n = len(self.y_train)
        self.val_n = len(self.y_val)
        self.test_n = len(self.y_test)
        logging.info(f"Training length = {self.n}")
        logging.info(f"Validation length = {self.val_n}")
        logging.info(f"Test length = {self.test_n}")
        logging.info("Loaded labels")

    def verify_image_quantity(self):
        self.X_base_train_paths = list(self.train_path.rglob(f"X_{self.single_dir_label}.*.npy"))
        self.X_base_val_paths = list(self.val_path.rglob(f"X_{self.single_dir_label}.*.npy"))
        self.X_base_test_paths = list(self.test_path.rglob(f"X_{self.single_dir_label}.*.npy"))
        if self.limit < 0 and self.n != len(self.X_base_train_paths):
            logging.error(f"# of training labels = {self.n}, # of training base images = {len(self.X_base_train_paths)}")
            raise Exception("Non matching lengths")
        if self.limit < 0 and self.val_n != len(self.X_base_val_paths):
            logging.error(f"# of validation labels = {self.val_n}, # validation base images {len(self.X_base_val_paths)}")
            raise Exception("Non matching lengths")
        if self.limit < 0 and self.test_n != len(self.X_base_test_paths):
            logging.error(f"# of test labels = {self.test_n}, # test base images {len(self.X_base_test_paths)}")
            raise Exception("Non matching lengths")
        if self.is_siamese:
            self.X_long_train_paths = list(self.train_path.rglob(f"X_long.*.npy"))
            self.X_long_val_paths = list(self.val_path.rglob(f"X_long.*.npy"))
            self.X_long_test_paths = list(self.test_path.rglob(f"X_long.*.npy"))
            if self.limit < 0 and self.n != len(self.X_long_train_paths):
                logging.error(f"# of training labels = {self.n}, # of training longitudinal images = {len(self.X_long_train_paths)}")
                raise Exception("Non matching lengths")
            if self.limit < 0 and self.val_n != len(self.X_long_val_paths):
                logging.error(f"# of validation labels = {self.val_n}, # of validation longitudinal images = {len(self.X_long_val_paths)}")
                raise Exception("Non matching lengths")
            if self.limit < 0 and self.test_n != len(self.X_long_test_paths):
                logging.error(f"# of testlabels = {self.test_n}, # of test longitudinal images = {len(self.X_long_test_paths)}")
                raise Exception("Non matching lengths")

    def define_datasets(self):
        self.X_base = []
        self.X_base_val = []
        self.X_base_test = []
        if self.is_siamese:
            self.X_long = []
            self.X_long_val = []
            self.X_long_test = []
            output_types = ({"left_input": tf.float32, "right_input": tf.float32}, tf.int8)
            output_shapes = ({"left_input": tf.TensorShape(self.input_shape), "right_input": tf.TensorShape(self.input_shape)}, tf.TensorShape([]))
        else:
            output_types = ({'single_input': tf.float32}, tf.int8)
            output_shapes = ({"single_input": tf.TensorShape(self.input_shape)}, tf.TensorShape([]))

        self.X_train_data = Dataset.from_generator(self.data_X_generator, 
            output_types=output_types,
            output_shapes=output_shapes)
        self.X_train_data = self.X_train_data.map(
            self.augment_map_siamese if self.is_siamese else self.augment_map_single,
            num_parallel_calls=tf.data.experimental.AUTOTUNE, deterministic=False)
        self.X_train_data = self.X_train_data.batch(self.batch_size, drop_remainder=True)

        self.X_val_data = Dataset.from_generator(self.data_X_val_generator,
            output_types=output_types,
            output_shapes=output_shapes)
        self.X_val_data = self.X_val_data.batch(self.batch_size, drop_remainder=True)

        self.X_test_data = Dataset.from_generator(self.data_X_test_generator,
            output_types=output_types,
            output_shapes=output_shapes)
        self.X_test_data = self.X_test_data.batch(self.batch_size, drop_remainder=True)

    def data_X_generator(self):
        shuffle = np.arange(self.n)
        if len(self.X_base) == self.n and ((not self.is_siamese) or len(self.X_long) == self.n):
            np.random.shuffle(shuffle)
        for i in range(self.n):
            suffix = f"{i:04d}"
            if len(self.X_base) <= i:
                self.X_base.append(np.load(self.train_path / f"X_{self.single_dir_label}.{suffix}.npy", mmap_mode='r'))
            if self.is_siamese and len(self.X_long) <= i:
                self.X_long.append(np.load(self.train_path / f"X_long.{suffix}.npy", mmap_mode='r'))
            j = shuffle[i]
            baseline = self.X_base[j]
            label = self.y_train[j]
            if self.is_siamese:
                longitudinal = self.X_long[j]
                yield {"left_input": baseline, "right_input": longitudinal}, label
            else:
                yield {'single_input': baseline}, label
    
    def data_X_val_generator(self):
        for i in range(self.val_n):
            suffix = f"{i:04d}"
            if len(self.X_base_val) <= i:
                self.X_base_val.append(np.load(self.val_path / f"X_{self.single_dir_label}.{suffix}.npy", mmap_mode='r'))
            if self.is_siamese:
                if len(self.X_long_val) <= i:
                    self.X_long_val.append(np.load(self.val_path / f"X_long.{suffix}.npy", mmap_mode='r'))
                yield {"left_input": self.X_base_val[i], "right_input": self.X_long_val[i]}, self.y_val[i]
            else:
                yield {'single_input': self.X_base_val[i]}, self.y_val[i]

    def data_X_test_generator(self):
        for i in range(self.test_n):
            suffix = f"{i:04d}"
            if len(self.X_base_test) <= i:
                self.X_base_test.append(np.load(self.test_path / f"X_{self.single_dir_label}.{suffix}.npy", mmap_mode='r'))
            if self.is_siamese:
                if len(self.X_long_test) <= i:
                    self.X_long_test.append(np.load(self.test_path / f"X_long.{suffix}.npy", mmap_mode='r'))
                yield {"left_input": self.X_base_test[i], "right_input": self.X_long_test[i]}, self.y_test[i]
            else:
                yield {'single_input': self.X_base_test[i]}, self.y_test[i]

    def augment_map_siamese(self, inp, out):
        base = inp['left_input']
        longitudinal = inp['right_input']
        base, longitudinal = tf.py_function(func=self.augment,
            inp=[base, longitudinal], 
            Tout=[tf.float32, tf.float32])
        return {'left_input': base, 'right_input': longitudinal}, out

    def augment_map_single(self, inp, out):
        base = inp['single_input']
        base = tf.py_function(func=self.augment_single,
            inp= [base], 
            Tout=tf.float32)
        return {'single_input': base}, out
            
    def augment(self, baseline, longitudinal):
        # Rotate up to 5 degrees in either direction and any axis
        planes = [(0,1), (1,2), (0,2)]
        angle = np.random.randint(6)
        neg = np.random.randint(2)
        angle = -angle if neg==1 else angle
        plane = np.random.randint(3)
        # plane = 0 # Always rotate in axial plane
        baseline = rotate(baseline, angle, axes=planes[plane], reshape=False)
        longitudinal = rotate(longitudinal, angle, axes=planes[plane], reshape=False)

        # Randomly flip in sagittal axis (left/right)
        flip = np.random.randint(2)
        if flip == 1:
            baseline = np.flip(baseline, axis=0)
            longitudinal = np.flip(longitudinal, axis=0)
        return baseline, longitudinal

    def augment_single(self, baseline):
        # Rotate up to 5 degrees in either direction and any axis
        planes = [(0,1), (1,2), (0,2)]
        angle = np.random.randint(6)
        neg = np.random.randint(2)
        angle = -angle if neg==1 else angle
        plane = np.random.randint(3)
        baseline = rotate(baseline, angle, axes=planes[int(plane)], reshape=False)

        # Randomly flip in sagittal axis (left/right)
        flip = np.random.randint(2)
        if flip == 1:
            baseline = np.flip(baseline, axis=0)
        return baseline

    def build_model(self):
        if self.is_siamese:
            left_input = Input(self.input_shape, name="left_input")
            right_input = Input(self.input_shape, name="right_input")
            self.build_child()
            logging.info("Child model built")
            (left_child, left_grad) = self.child_model(left_input)
            (right_child, right_grad) = self.child_model(right_input) if not self.child_model2 else self.child_model2(right_input)
            (self.model, self.left_grad_model, self.right_grad_model) = self.siamese_join(left_input, right_input, left_child, right_child, left_grad, right_grad)
            self.right_grad_layer = right_grad
            self.left_grad_layer = left_grad
        else:
            single_input = Input(self.input_shape, name="single_input")
            self.build_child()
            (child, grad) = self.child_model(single_input)
            x = self.apply_tail(child)
            x = Dense(self.classes, 
                name='final')(x)
            x = Activation('sigmoid')(x)
            self.model = Model(inputs=single_input, outputs=x, name=self.name)
            self.grad_model = Model(inputs=single_input, outputs=[grad, x])
            self.grad_layer = grad
        plot_model(self.child_model, to_file=f"pngs/{self.name}_child.png", show_shapes=True)
        plot_model(self.model, to_file=f"pngs/{self.name}_fullmodel.png", show_shapes=True)
            
    def siamese_join(self, left_input, right_input, left_child, right_child, left_grad, right_grad):
        x = self.join_function(left_child, right_child)
        x = self.apply_tail(x)
        x = Dense(self.classes, 
            name='final')(x)
        x = Activation('sigmoid')(x)
        model = Model(inputs=[left_input,right_input], outputs=x, name=f"{self.name}_siamese")
        right_grad_model = Model(inputs=[left_input, right_input], outputs=[right_grad, x])
        left_grad_model = Model(inputs=[left_input, right_input], outputs=[left_grad, x])
        return model, left_grad_model, right_grad_model

    def join_function(self, left_child, right_child):
        return subtract([left_child, right_child])

    def build_child(self):
        pass

    def apply_tail(self, x):
        args = [x] + self.tail_args
        return self.tails[self.tail_name](*args)

    def compile_model(self):
        self.description = "\n"
        def capture_summary(summary_line):
            self.description += summary_line + "\n"
        self.child_description = "\n"
        def capture_child_summary(summary_line):
            self.child_description += summary_line + "\n"

        logging.info("Compiling model....")
        self.load_checkpoint()
        self.freeze_layers()
        def lr_metric(y_true, y_pred):
            return self.opt.lr
        self.model.compile(optimizer=self.opt, 
            loss='binary_crossentropy', 
            metrics=['FalseNegatives', 'FalsePositives', 'TrueNegatives', 'TruePositives', 'Precision', 'Recall', lr_metric, 'accuracy'])
        self.child_model.summary(print_fn=capture_child_summary, line_length=120)
        self.model.summary(print_fn=capture_summary, line_length=120)
        logging.info(self.child_description)
        logging.info(self.description)
        self.transfer_child()
        self.store_parameters()
    
    def store_parameters(self):
        self.parameters['name'] = self.name
        self.parameters['comment'] = self.comment
        self.parameters['class'] = type(self).__name__
        self.parameters['child_transfer'] = self.child_transfer
        self.parameters['input_shape'] = str(self.input_shape)
        self.parameters['train_dir'] = self.train_dir
        self.parameters['val_dir'] = self.val_dir
        self.parameters['xval'] = self.xval
        self.parameters['n'] = self.n
        self.parameters['val_n'] = self.val_n
        self.parameters['attempted_epochs'] = self.epochs
        self.parameters['batch_size'] = self.batch_size
        self.parameters['freeze_child'] = self.freeze_child
        self.parameters['unfreeze_end'] = self.unfreeze_end
        self.parameters['freeze_tail'] = self.freeze_tail
        self.parameters['prev_checkpoint'] = self.prev_checkpoint
        self.parameters['single_dir_label'] = self.single_dir_label
        self.parameters['model_config'] = json.dumps(self.model.get_config())
        self.parameters['model_description'] = self.description
        self.parameters['child_model_config'] = json.dumps(self.model.get_config())
        self.parameters['child_description'] = self.child_description
        self.parameters['optimizer'] = self.opt.get_config().get('name')
        self.parameters['learning_rate'] = self.opt.get_config().get('learning_rate')
        self.parameters['optimizer_config'] = json.dumps({key:str(value) for key,value in self.opt.get_config().items()})
        self.parameters['lr_start'] = self.lr_start
        self.parameters['lr_end'] = self.lr_end
        # self.parameters['optimizer_confif'] = json.dumps(self.opt.get_config())
        self.parameters['stages'] = str(self.stages) if hasattr(self,'stages') else ""
        self.parameters['filters'] = str(self.filters) if hasattr(self, 'filters') else ""
        self.parameters['tail'] = self.tail_name
        self.parameters['tail_args'] = str(self.tail_args)

    def fit_model(self):
        self.attempt += 1
        self.plot.attempt = self.attempt
        self.store_parameters()
        self.plot.parameters = self.parameters
        logging.info(f"Training {self.epochs} epochs with patience = {self.patience}...")
        self.plot.add_epochs(self.epochs)
        steps_per_epoch = self.n // self.batch_size
        validation_steps = self.val_n // self.batch_size
        self.checkpoint_filename = self.name if self.attempt==1 else f"{self.name}_{str(self.attempt)}"
        self.checkpoint = ModelCheckpoint(f"checkpoints/{self.checkpoint_filename}.tf",
                                            save_weights_only=True,
                                            save_best_only=True,
                                            period=1,
                                            verbose=1)
        self.stopping = EarlyStopping(monitor='val_loss',
                                        patience=self.patience)
        callbacks = [self.plot, self.logger, self.checkpoint, self.stopping]
        if self.lr_end > 0:
            def schedule(epoch, lr):
                #return (self.lr_start-self.lr_end)*np.exp(-5*epoch/self.epochs)+self.lr_end
                new_lr = self.lr_start*((self.lr_end/self.lr_start)**(epoch/self.epochs))
                #logging.info(f"Epoch #{epoch} learning rate = {new_lr}")
                return new_lr
            self.lr_sched = LearningRateScheduler(schedule, verbose=0)
            callbacks.append(self.lr_sched)
        self.model.fit(self.X_train_data.repeat(self.epochs).prefetch(1), 
            epochs=self.epochs,
            validation_data=self.X_val_data.repeat(self.epochs), 
            callbacks=callbacks, 
            steps_per_epoch=steps_per_epoch,
            validation_steps=validation_steps)
        self.save_child()
        logging.info("Training session complete")

    def reload_best_weights(self):
        logging.info()

    def save_child(self):
        filename = f"child_saves/{self.name}_child_weights.h5"
        self.child_model.save_weights(filename)
        logging.info(f"Saved child weights at {filename}")
        logging.info(f"Use only '{self.name}' as the child_transfer name in the future")

    def transfer_child(self):
        if len(self.child_transfer) > 0:
            filename = f"child_saves/{self.child_transfer}_child_weights.h5"
            self.child_model.load_weights(filename)
            logging.info(f"Transferred child weights from {filename}")
            if self.child_model2:
                self.child_model2.load_weights(filename)
                logging.info(f"Transferred child weights from {filename} to second twin")
        else:
            logging.info("No child weights transferred")

    def load_checkpoint(self):
        if len(self.prev_checkpoint) > 0:
            filename = f"checkpoints/{self.prev_checkpoint}"
            filename = filename if filename[-2:] == 'h5' else f"{filename}.tf"
            self.model.load_weights(filename)
            logging.info(f"Loaded checkpoint from {filename}")
        else:
            logging.info("No checkpoint loaded")

    def freeze_layers(self):
        if self.freeze_child:
            logging.info("Freezing child layers")
            for layer in self.child_model.layers:
                layer.trainable = False
            if self.unfreeze_end:
                logging.info("Unfreezing last child layers")
                past_first = False
                for layer in reversed(self.child_model.layers):
                    layer.trainable = True
                    if layer.name[:3] == "add" or layer.name[:3] == "max":
                        if past_first:
                            break
                        else:
                            past_first = True
            if self.child_model2:
                logging.info("Freezing second child layers")
                for layer in self.child_model2.layers:
                    layer.trainable = False
                if self.unfreeze_end:
                    logging.info("Unfreezing second child layers")
                    past_first = False
                    for layer in reversed(self.child_model2.layers):
                        layer.trainable = True
                        if layer.name[:3] == "add" or layer.name[:3] == "max":
                            if past_first:
                                break
                            else:
                                past_first = True
        else:
            logging.info("Child layers not frozen")

        if self.freeze_tail:
            logging.info("Freezing tail layers")
            for layer in self.model.layers:
                layer.trainable = False
            logging.info("Unfreeze up until last dense layer")
            for layer in reversed(self.model.layers):
                layer.trainable = True
                if layer.name == 'final':
                    break
        else:
            logging.info("Tail layers not frozen")

    def analyze_3d(self):
        logging.info("Validation metrics:")
        return self._print_analysis(self.val_n, self.X_val_data, self.y_val)
        #self.plot.final_plot()

    def analyze_test(self):
        logging.info("Test metrics:")
        return self._print_analysis(self.test_n, self.X_test_data, self.y_test)

    def show_pred_test(self):
        steps = self.test_n // self.batch_size
        n = steps * self.batch_size
        label_class = End_Data_Volume(self.y_test[:n])
        y_true = label_class.y_end
        pred = self.model.predict(self.X_test_data, steps=steps)
        pred_class = End_Data_Volume(pred)
        y_pred = pred_class.y_end
        return y_true, y_pred

    def _print_analysis(self, n_full, X, y):
        steps = n_full // self.batch_size
        n = steps * self.batch_size
        label_class = End_Data_Volume(y[:n])
        y_true = label_class.y_end
        pred = self.model.predict(X, steps=steps)
        auc = roc_auc_score(y_true, pred)
        pred_class = End_Data_Volume(pred)
        y_pred = pred_class.y_end
        logging.info("\n"+classification_report(y_true, y_pred))
        logging.info(f"Balanced accuracy = {balanced_accuracy_score(y_true, y_pred):.2f}")
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
        sens = tp/(tp+fn)
        spec = tn/(tn+fp)
        logging.info(f"True negatives: {tn}")
        logging.info(f"False positives: {fp}")
        logging.info(f"False negatives: {fn}")
        logging.info(f"True positives: {tp}")
        logging.info(f"Specificity = {spec:.2f}")
        logging.info(f"Sensitivity = {sens:.2f}")
        ba = (sens+spec)/2
        logging.info(f"Calculated bal accuracy = {ba:.2f}")
        logging.info(f"AUC: {auc:.2f}")
        return ba, auc
        
    def report_analysis(self):
        # plot_model(self.model, to_file="resnet_siamese.png", show_shapes=True)
        report = classification_report(self.val.y_end, self.pred.y_end, output_dict=True)
        
        if not isfile("training_results.csv"):
            training_results = pd.DataFrame(columns=["title", "description", "epochs", "learning_rate", "features", "metric", "label", "value"])
        else:
            training_results = pd.read_csv("training_results.csv")

        title = "3D longitudinal siamese resnet"
        for label, value in report.items():
            row = {"title": title, "description": self.description, "epochs": self.epochs, "learning_rate": self.learning_rate, "features": 0}
            if label == "accuracy":
                row["metric"] = label
                row["label"] = None
                row["value"] = value
                training_results = training_results.append(row, ignore_index=True)
            else:
                for metric, score in value.items():
                    row["metric"] = metric
                    row["label"] = label
                    row["value"] = score
                    training_results = training_results.append(row, ignore_index=True)

        training_results.to_csv(f"training_results_{self.name}.csv", index=False)

    def get_heatmaps(self, left=True):
        heatmaps = []
        for i in range(len(self.X_base_test)):
            self.grad_cam(i, left)
            logging.info(i)
            max_value = self.cam.max()
            min_value = self.cam.min()
            diff = max_value - min_value
            factor = 1000.0/diff
            heatmap = factor*(self.cam-min_value)
            if not np.isnan(heatmap.max()):
                heatmaps.append(heatmap)
        self.heatmaps = np.array(heatmaps)
        self.avg_heatmap = self.heatmaps.mean(axis=0)
        zoom_factor = [input_dim/self.avg_heatmap.shape[axis] for axis, input_dim in enumerate(self.input_shape[:3])]
        self.avg_heatmap = zoom(self.avg_heatmap, zoom_factor)

    def grad_cam(self, sample=0, left=True, zoom_now = False):
        # for layer in reversed(self.model.layers):
        #     if len(layer.output_shape) == 5:
        #         last_layer = layer.name
        #         break
        # logging.info(f"Layer to output: {last_layer}")
        # grad_model = Model(
        #     inputs=[self.model.inputs],
        #     outputs=[self.model.get_layer(last_layer).output, self.model.output])
        # grad = self.model.inputs

        # logging.info("loaded grad_cam")
        if self.is_siamese:
            if left:
                self.grad_model = self.left_grad_model
                self.grad_layer = self.left_grad_layer
            else:
                self.grad_model = self.right_grad_model
                self.grad_layer = self.right_grad_layer
        with tf.GradientTape() as tape:
            tape.watch(self.grad_layer)
            if self.is_siamese:
                inputs = [tf.cast(np.expand_dims(self.X_base_test[sample], axis=0), tf.float32),
                            tf.cast(np.expand_dims(self.X_long_test[sample], axis=0), tf.float32)]
            else:
                inputs = tf.cast(np.expand_dims(self.X_base_test[sample], axis=0), tf.float32)
            logging.info("achieved cast")
            (conv_outputs,predictions) = self.grad_model(inputs)
            logging.info("got predictions")
            loss = predictions[:,0]

        grads = tape.gradient(loss, conv_outputs)
        logging.info("got grads")
        self.conv_outputs = conv_outputs
        self.predictions = predictions
        self.loss = loss
        self.grads = grads
        cast_conv_outputs = tf.cast(conv_outputs > 0, "float32")
        cast_grads = tf.cast(grads > 0, "float32")

        guided_grads = cast_conv_outputs * cast_grads * grads
        #guided_grads = grads

        conv_outputs = conv_outputs[0]
        guided_grads = guided_grads[0]
        weights = tf.reduce_mean(guided_grads, axis=(0, 1, 2))
        cam = tf.reduce_sum(tf.multiply(weights, conv_outputs), axis=-1)
        logging.info(f"Cam shape = {str(cam.shape)}")
        self.cam = cam.numpy()
        if zoom_now:
            zoom_factor = [input_dim/cam.shape[axis] for axis, input_dim in enumerate(self.input_shape[:3])]
            self.heatmap = zoom(self.cam, zoom_factor)

class ResNet3D(BaseModel):

    def __init__(self, stages, filters, name, comment, compile_now=True):
        self.stages = stages
        self.filters = filters
        super().__init__(name, comment, compile_now)

    def set_parameters(self):
        super().set_parameters()
        self.train_dir = 'train_ad_cn_2mm'
        self.val_dir = 'val_ad_cn_2mm'
        self.reg = 0.0001
        self.bn_eps = 2e-5
        self.bn_mom = 0.9
        self.initializer = 'he_normal'
        self.batch_size = 12
        self.learning_rate = 1e-4
        # self.opt = SGD(lr=self.learning_rate, nesterov=True)
        self.opt = Adam(lr=self.learning_rate, epsilon=1e-8)
        self.tail_name = "long_2048"
        #self.tail_args = [(2,2,2)]
        self.tail_args = []
        self.single_dir_label = "base"
        self.child_transfer = ""
        self.freeze_child = False
        self.patience = 100
        self.prev_checkpoint = ""

    def build_child(self):
        inputs = Input(shape=self.input_shape, name='resnet_input')
        end = self.resnet_child(inputs)
        self.child_model = Model(inputs=inputs, outputs=end, name="ResNet3D")

    def resnet_child(self, inputs):
        x = BatchNormalization(epsilon=self.bn_eps, momentum=self.bn_mom)(inputs)
        x = Conv3D(self.filters[0], (5, 5, 5), use_bias=False, padding="same", 
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(self.reg))(x)

        for i in range(len(self.stages)):
            stride = (1, 1, 1) if i == 0 else (2, 2, 2)
            x = self.residual_module(x, self.filters[i + 1], stride, shrink=True)
            for _ in range(self.stages[i] - 1):
                x = self.residual_module(x, self.filters[i + 1], (1, 1, 1))
            if i==3:
                grad = Activation('linear', name='grad')(x)
                self.grad_layer = grad

        x = BatchNormalization(epsilon=self.bn_eps, momentum=self.bn_mom)(x)
        relu = Activation('relu', name='rn_relu_final')(x)
        return relu, grad

    def residual_module(self, data, K, stride, shrink=False):
        shortcut = data

        bn1 = BatchNormalization(epsilon=self.bn_eps, momentum=self.bn_mom)(data)
        act1 = Activation("relu")(bn1)
        conv1 = Conv3D(K // 4, (1, 1, 1), 
            use_bias=False,
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(self.reg))(act1)

        bn2 = BatchNormalization(epsilon=self.bn_eps, momentum=self.bn_mom)(conv1)
        act2 = Activation("relu")(bn2)
        conv2 = Conv3D(K // 4, (3, 3, 3), strides=stride, 
            kernel_initializer=self.initializer,
            padding="same", use_bias=False, kernel_regularizer=l2(self.reg))(act2)

        bn3 = BatchNormalization(epsilon=self.bn_eps, momentum=self.bn_mom)(conv2)
        act3 = Activation("relu")(bn3)
        conv3 = Conv3D(K, (1, 1, 1), use_bias=False, 
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(self.reg))(act3)

        # if we are to shrink the spatial size (shrink=True), apply a CONV layer to the shortcut
        if shrink:
            shortcut = Conv3D(K, (1, 1, 1), strides=stride, use_bias=False, 
                kernel_initializer=self.initializer,
                kernel_regularizer=l2(self.reg))(act1)

        # add together the shortcut and the final CONV
        x = add([conv3, shortcut])
        return x

class ResNet_Siamese(ResNet3D):

    is_siamese = True

    def set_parameters(self):
        super().set_parameters()
        self.child_transfer = "resnet_deep_transfer_002"
        self.prev_checkpoint = ""

class ResNetConcat(ResNet_Siamese):

    def join_function(self, left_child, right_child):
        return concatenate([left_child, right_child])

class ResNetTwin(ResNetConcat):

    def set_parameters(self):
        super().set_parameters()
        self.patience = 20
        # self.unfreeze_end = True
        # self.opt = Adam(lr=self.learning_rate, epsilon=1e-8)
        # self.tail_name = "short_128"
        # self.tail_args = [(12,14,12)]

    def build_child(self):
        inputs = Input(shape=self.input_shape, name='resnet_input')
        end = self.resnet_child(inputs)
        self.child_model = Model(inputs=inputs, outputs=end, name="ResNet3D")
        inputs2 = Input(shape=self.input_shape, name='resnet_input2')
        end2 = self.resnet_child(inputs2)
        self.child_model2 = Model(inputs=inputs2, outputs=end2, name="ResNet3D_2")

class WideResidual(BaseModel):

    def __init__(self, filters, name, comment, compile_now=True):
        # self.stages = stages
        self.filters = filters
        super().__init__(name, comment, compile_now)

    def set_parameters(self):
        super().set_parameters()
        self.train_dir = 'train_ad_cn_2mm'
        self.val_dir = 'val_ad_cn_2mm'
        self.reg = 0.0001
        self.bn_eps = 2e-5
        self.bn_mom = 0.9
        self.initializer = 'he_normal'
        self.batch_size = 12
        self.learning_rate = 1e-4
        # self.opt = SGD(lr=self.learning_rate, nesterov=True)
        self.opt = Adam(lr=self.learning_rate, epsilon=1e-8)
        self.tail_name = "long_2048"
        #self.tail_args = [(2,2,2)]
        self.tail_args = []
        self.single_dir_label = "base"
        self.child_transfer = ""
        self.freeze_child = False
        self.patience = 100
        self.prev_checkpoint = ""

    def build_child(self):
        inputs = Input(shape=self.input_shape, name='resnet_input')
        end = self.resnet_child(inputs)
        self.child_model = Model(inputs=inputs, outputs=end, name="WideResidual")

    def resnet_child(self, inputs):
        x = inputs
        for i, filters in enumerate(self.filters):
            x = self.residual_module(x, filters, 2)
            if i==1:
                grad = Activation('linear', name='grad')(x)
                self.grad_layer = grad
        relu = x
        return relu, grad

    def residual_module(self, data, K, stride, dropout=True):
        shortcut = Conv3D(K, (1, 1, 1), strides=stride,
            padding='same',
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(self.reg))(data)

        conv1 = Conv3D(K, 3, 
            strides = stride,
            padding='same',
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(self.reg))(data)
        bn1 = BatchNormalization()(conv1)
        act1 = Activation("relu")(bn1)
        if dropout:
            act1 = SpatialDropout3D(0.3)(act1)
        conv2 = Conv3D(K, 3,
            kernel_initializer=self.initializer,
            padding="same", kernel_regularizer=l2(self.reg))(act1)

        # add together the shortcut and the final CONV
        x = add([conv2, shortcut])
        x = BatchNormalization()(x)
        x = Activation('relu')(x)
        return x

class WideResidualSiamese(WideResidual):

    is_siamese = True
    
class WideResidualConcat(WideResidualSiamese):

    def join_function(self, left_child, right_child):
        return concatenate([left_child, right_child])

class WideResidualTwin(WideResidualConcat):

    def build_child(self):
        inputs = Input(shape=self.input_shape, name='wrn_input')
        end = self.resnet_child(inputs)
        self.child_model = Model(inputs=inputs, outputs=end, name="WideResidual")
        inputs2 = Input(shape=self.input_shape, name='wrn_input2')
        end2 = self.resnet_child(inputs2)
        self.child_model2 = Model(inputs=inputs2, outputs=end2, name="WideResidual_2")

class Sequential3D(BaseModel):

    def __init__(self, filters, name, comment, compile_now=True):
        self.filters = filters
        super().__init__(name, comment, compile_now)

    def set_parameters(self):
        super().set_parameters()
        self.train_dir = 'train_mmse_2mm'
        self.val_dir = 'val_mmse_2mm'
        self.learning_rate = 1e-5
        self.opt = Adam(self.learning_rate,epsilon=1e-8)
        self.batch_size = 12
        self.reg = 0.005
        self.bn_eps = 2e-5
        self.bn_mom = 0.9
        self.initializer = 'he_normal'
        self.tail_name = "long_2048"
        self.child_transfer = ""

    def build_child(self):
        inputs = Input(shape=self.input_shape, name='sequential_input')
        end = self.sequential_child(inputs)
        self.child_model = Model(inputs=inputs, outputs=end, name="sequential_child")

    def sequential_child(self, inputs):
        x = inputs
        for i, filters in enumerate(self.filters):
            logging.info(f"Building stage with {filters} filters")
            x, last = self.sequential_segment(x, filters)
            if i==2:
                grad = Activation('linear', name='grad')(last)
                # self.grad_layer = grad
        return x, grad

    def sequential_segment(self, x, filters, pool=True):
        logging.info("conv step...")
        conv = Conv3D(filters, 3, strides=1,
                    padding='same',
                    kernel_initializer=self.initializer,
                    kernel_regularizer=l2(self.reg))(x)
        logging.info('bn step...')
        x = BatchNormalization(momentum=0.99)(conv)
        act = Activation('relu')(x)
        x = MaxPool3D(2, strides=2)(act)
        return x, conv

class SequentialSiamese(Sequential3D):

    is_siamese = True

    def set_parameters(self):
        super().set_parameters()
        self.opt = Adam(self.learning_rate,epsilon=1e-8)

class SequentialConcat(SequentialSiamese):

    def join_function(self, left_child, right_child):
        return concatenate([left_child, right_child])

    def set_parameters(self):
        super().set_parameters()
        self.tail_name = "medium_1300"

class SequentialTwin(SequentialConcat):

    def build_child(self):
        inputs = Input(shape=self.input_shape, name='sequential_input')
        end = self.sequential_child(inputs)
        self.child_model = Model(inputs=inputs, outputs=end, name="Sequential3D")
        inputs2 = Input(shape=self.input_shape, name='sequential_input2')
        end2 = self.sequential_child(inputs2)
        self.child_model2 = Model(inputs=inputs2, outputs=end2, name="Sequential3D_2")

class Classic(BaseModel):

    def set_parameters(self):
        super().set_parameters()
        self.train_dir = 'train_mmse_2mm'
        self.val_dir = 'val_mmse_2mm'
        self.learning_rate = 1e-4
        self.opt = Adam(lr=self.learning_rate, beta_1=0.9, beta_2=0.999, epsilon=None, decay=0.000000001, amsgrad=False)
        self.tail_name = "long_2048"
        self.child_transfer = "classic_ad_cn_004"
        self.batch_size = 8

    def build_child(self):
        reg = 0.005
        inputs = Input(shape=self.input_shape, name='classic_input')

        x = BatchNormalization(momentum=0.99)(inputs)

        x = Conv3D(16, (3,3,3), use_bias=False, padding='same', 
                    kernel_regularizer=l2(reg))(x)
        x = BatchNormalization(momentum=0.99)(x)
        x = LeakyReLU(alpha=0.01)(x)
        x = AveragePooling3D(pool_size=3, strides=2, padding='same')(x)

        x = Conv3D(32, (3,3,3), use_bias=False, padding='same', 
                    kernel_regularizer=l2(reg))(x)
        x = BatchNormalization(momentum=0.99)(x)
        x = LeakyReLU(alpha=0.01)(x)
        x = AveragePooling3D(pool_size=3, strides=2, padding='same')(x)

        x = Conv3D(32, (3,3,3), use_bias=False, padding='same', 
                    kernel_regularizer=l2(reg))(x)
        x = BatchNormalization(momentum=0.99)(x)
        x = LeakyReLU(alpha=0.01)(x)
        x = AveragePooling3D(pool_size=3, strides=2, padding='same')(x)

        x = Conv3D(32, (3,3,3), use_bias=False, padding='same', 
                    kernel_regularizer=l2(reg))(x)
        x = BatchNormalization(momentum=0.99)(x)
        x = LeakyReLU(alpha=0.01, name='child_lrelu_final')(x)

        self.child_model = Model(inputs=inputs, outputs=x, name='classic')

class ClassicSiamese(Classic):

    is_siamese = True

class ClassicConcat(ClassicSiamese):

    def join_function(self, left_net, right_net):
        return concatenate([left_net, right_net])

class Attention(BaseModel):

    def __init__(self, filters, name, comment, compile_now=True):
        self.filters = filters
        super().__init__(name, comment, compile_now)

    def set_parameters(self):
        super().set_parameters()
        self.train_dir = 'train_ad_cn_2mm'
        self.val_dir = 'val_ad_cn_2mm'
        self.reg = 0.0001
        self.bn_eps = 2e-5
        self.bn_mom = 0.9
        self.initializer = 'he_normal'
        self.batch_size = 12
        self.learning_rate = 1e-4
        # self.opt = SGD(lr=self.learning_rate, nesterov=True)
        self.opt = Adam(lr=self.learning_rate, epsilon=1e-8)
        self.tail_name = "short_128"
        self.tail_args = [(2,2,2)]
        # self.tail_args = []
        self.single_dir_label = "base"
        self.child_transfer = ""
        self.freeze_child = False
        self.patience = 100
        self.prev_checkpoint = ""

    def build_child(self):
        inputs = Input(shape=self.input_shape, name='attention_input')
        end = self.attention_child(inputs)
        self.child_model = Model(inputs=inputs, outputs=end, name="Attention")

    def attention_block(self, filters, data):
        x = Conv3D(filters, 3, 1, 
            padding='same', 
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(self.reg))(data)
        x = Activation('relu')(x)
        x = multiply([x, data])
        return x

    def residual_block(self, filters, stride, data):
        shortcut = Conv3D(filters, 1, stride,
            padding='same',
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(0.01))(data)
        x = Conv3D(filters, 1, 1,
            padding='same',
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(0.01))(data)
        x = Activation('relu')(x)
        x = Conv3D(filters, 3, stride,
            padding='same',
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(0.01))(x)
        x = Activation('relu')(x)
        x = add([shortcut, x])
        x = BatchNormalization(momentum=0.99)(x)
        return x
        
    def attention_child(self, inputs):
        x = BatchNormalization(epsilon=self.bn_eps, momentum=self.bn_mom)(inputs)
        x = Conv3D(self.filters[0], (7, 7, 7), strides = 2, 
            use_bias=False, 
            padding="same", 
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(self.reg))(x)
        x = Activation('relu')(x)
        x = MaxPool3D(2, 2, padding='same')(x)
        x = Conv3D(self.filters[0], (3, 3, 3), use_bias=False, padding="same", 
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(self.reg))(x)
        x = Activation('relu')(x)
        x = Conv3D(self.filters[0], (3, 3, 3), use_bias=False, padding="same", 
            kernel_initializer=self.initializer,
            kernel_regularizer=l2(self.reg))(x)
        x = Activation('relu')(x)
        x = self.residual_block(self.filters[0], 2, x)
        x = self.residual_block(self.filters[1], 1, x)
        x = self.residual_block(self.filters[1], 2, x)
        x = self.attention_block(self.filters[1], x)
        x = self.residual_block(self.filters[2], 1, x)
        x = self.residual_block(self.filters[2], 2, x)
        x = self.attention_block(self.filters[2], x)
        x = self.residual_block(self.filters[3], 1, x)
        x = self.residual_block(self.filters[3], 2, x)

        return x

class Tails:

    @staticmethod
    def short_128(x, pool_filter):
        x = MaxPool3D(pool_filter)(x)
        x = Flatten()(x)
        x = Dense(128, kernel_regularizer=l2(0.005))(x)
        x = Activation('relu')(x)
        return x

    @staticmethod
    def long_2048(x):
        x = BatchNormalization(momentum=0.99)(x)
        x = LeakyReLU(alpha=0.01)(x)
        x = Dropout(0.5)(x)
        x = Flatten()(x)

        x = Dense(2048, 
            kernel_regularizer=l2(0.005))(x)
        x = BatchNormalization(momentum=0.99)(x)
        x = LeakyReLU(alpha=0.01)(x)
        x = Dropout(0.5)(x)

        x = Dense(1024, 
            kernel_regularizer=l2(0.005))(x)
        x = BatchNormalization(momentum=0.99)(x)
        x = LeakyReLU(alpha=0.01)(x)

        x = Dropout(0.5)(x)
        return x

    @staticmethod
    def medium_1300(x):
        drop = Dropout(0.5)(x)
        flat = Flatten()(drop)
        fc1 = Dense(1300,
                    kernel_regularizer=l2(0.005))(flat)
        relu1 = Activation('relu')(fc1)
        dense2 = Dense(50,
                    kernel_regularizer=l2(0.005))(relu1)
        relu2 = Activation('relu')(dense2)
        return relu2

    @staticmethod
    def gap(x):
        x = GlobalAveragePooling3D()(x)
        return x

    @staticmethod
    def gmp(x):
        x = GlobalMaxPooling3D()(x)
        return x

    @staticmethod
    def flat(x):
        x = Flatten()(x)
        return x

    @staticmethod
    def flat128(x):
        x = Flatten()(x)
        x = Dense(128, kernel_regularizer=l2(0.005))(x)
        x = Activation('relu')(x)
        x = Dropout(0.5)(x)
        return x
